unit filetools;

interface


uses sysutils, tools, utypes, FileType;

type
     EStreamError=class(Exception);

     TStreamWrite=class
     public
      function GetPosition:LongWord; virtual; abstract;
      procedure AddBit(Value:Boolean); virtual; abstract;
      procedure AddArrayOfBit(const value:array of byte;NbBit:LongWord);  virtual; abstract;
      procedure AddByte(Value:Byte); virtual; abstract;
      procedure AddBuffer(var Value;NbByte:LongWord); virtual; abstract;
     end;

     EFileStreamError=class(EStreamError);
     type TTabBuffer=array[0..MaxInt-1] of byte;
     PTabBuffer=^TTabBuffer;

     TStreamWriteFile=class(TStreamWrite)
     private
      F:File;
      Filename:String;
      FileOpened:Boolean;
      BufferSize,BufferPointer,BufferSizeBit:LongWord;
      Buffer:PTabBuffer;
      procedure flush;
     public
      function GetPosition:LongWord;  override;
      procedure AddBit(Value:Boolean); override;
      procedure AddArrayOfBit(const value:array of byte;NbBit:LongWord); override;
      procedure AddByte(Value:Byte); override;
      procedure AddBuffer(var Value;NbByte:LongWord); override;

      constructor Create(FileName:String;BufferSize:LongWord);
      procedure Init;
      procedure Close;
      destructor Destroy; override;
     end;

     TStreamRead=class
     public
      function getPosition:LongWord; virtual; abstract;
      procedure setPosition(Position:LongWord); virtual; abstract;
      function ReadBit:Integer; virtual; abstract;
      function ReadByte:Byte; virtual; abstract;
      procedure ReadBuffer(var Value;NbByte:LongWord); virtual; abstract;
     end;

     TStreamReadFile=class(TStreamRead)
     private
      F:File;
      Filename:String;
      FileOpened:Boolean;
      BufferSize,BufferPointer,BufferFilled:LongWord;
      Buffer:PTabBuffer;
      procedure read;
     public
      function getPosition:LongWord; override;
      procedure setPosition(Position:LongWord); override;
      function ReadBit:Integer; override;
      function ReadByte:Byte; override;
      procedure ReadBuffer(var Value;NbByte:LongWord); override;

      constructor Create(FileName:String;BufferSize:LongWord);
      procedure Init;
      procedure Close;
      destructor Destroy; override;
     end;






     procedure HuffmanDiffCompression(LargX,LargY:Integer;TypeData:Integer;Image:Pointer;OutPutStream:TStreamWrite;ProcessProgress:TProcessProgress);

type EDataExtractionError=class(ExceptionCode);

     procedure HuffmanDiffUnCompression(InputStream:TStreamRead;LargX,LargY:Integer;TypeData:Integer;Image:Pointer;X1,Y1,X2,Y2:Integer;ProcessProgress:TProcessProgress);



type EErrorLoadFile=class(ExceptionCode);
     EErrorSaveFile=class(ExceptionCode);
     
     procedure LoadCPA(const Name:PChar;
                       var HighValue,LowValue:TThreshold;var Infos_Image:TInfos_Image;
                       var SizeX,SizeY:Integer;var TypeData:Integer;var Image:Pointer;var NbrePlan:Integer;
                       ProcessProgress:TProcessProgress);

     procedure WriteCPA(const Name:string;var SeuilHaut,SeuilBas;var Infos_Image:TInfos_Image;
                        LargX,LargY:Integer;TypeData:Integer;Image:Pointer;NbrePlan:Byte;
                        ProcessProgress:TProcessProgress);
                  

implementation


uses Geredata;

{const
S_8579=' Unable to read the image (missing or corrupted data)';
S_8580='Unknown compressing method !';
S_8581='Corrupted data';
S_8582=' (bad Huffman header)';
S_8583='Data corrupted (bad Huffman header)';
S_8584='Impossible to read the Huffman header';
S_8585='Bad file signature';}



constructor TStreamWriteFile.Create(FileName:String;BufferSize:LongWord);
begin
 inherited Create;
 Self.FileName:=FileName;
 Self.BufferSize:=BufferSize;
 BufferSizeBit:=BufferSize shl 3;
 Buffer:=Nil;
 FileOpened:=False;
end;

procedure TStreamWriteFile.Init;
begin
 if FileOpened then
  Close;
 assignFile(F,FileName);
 try
 rewrite(F,1);
 FileOpened:=True;
 if Buffer=Nil then
  getmem(Buffer,BufferSize);
 fillchar(Buffer^,BufferSize,0);
 BufferPointer:=0;
 except
  on E:EOutOfMemory do
   raise EFileStreamError.Create('Out of memory !');
  on E:EInOutError do
   raise EFileStreamError.Create(E.Message);
 end;
end;

procedure TStreamWriteFile.Close;
begin
 flush;
 if FileOpened then
 begin
  closeFile(F);
  FileOpened:=False;
 end;
end;

procedure TStreamWriteFile.flush;
begin
 try
 if BufferPointer>0 then
 begin
  BlockWrite(F,Buffer^,((BufferPointer+7) shr 3));
  BufferPointer:=0;
  fillchar(Buffer^,BufferSize,0);
 end;
 except
  on E:EInOutError do
   raise EFileStreamError.Create(E.Message);
 end;
end;

function TStreamWriteFile.GetPosition:LongWord;
var FileP:LongWord;
begin
 try
 FileP:=FilePos(F);
 result:=FileP+BufferPointer;
 except
  on E:EInOutError do
   raise EFileStreamError.Create(E.Message);
 end;
end;

procedure TStreamWriteFile.AddBit(Value:Boolean);
var BytePos,BitPos:LongWord;
begin
 if BufferPointer>=BufferSizeBit then
  flush;
 if Value then
 begin
  BytePos:=BufferPointer shr 3;
  BitPos:=BufferPointer and $00000007;
  Buffer[BytePos]:=Buffer[BytePos] or (1 shl BitPos);
 end;          
 inc(BufferPointer);
end;

procedure TStreamWriteFile.AddArrayOfBit(const value:array of byte;NbBit:LongWord);
var I:LongWord;
begin                               
 for I:=0 to NbBit-1 do
 begin
  if BufferPointer>=BufferSizeBit then
   flush;
  if boolean((Value[I shr 3] shr (I and $00000007)) and 1) then
   Buffer[BufferPointer shr 3]:=Buffer[BufferPointer shr 3] or (1 shl (BufferPointer and $00000007));
  inc(BufferPointer);
 end;
end;

procedure TStreamWriteFile.AddByte(Value:Byte);
begin
 BufferPointer:=(BufferPointer+7) and $FFFFFFF8;
 if BufferPointer>=BufferSizeBit then
  flush;
 Buffer[BufferPointer shr 3]:=Value;
 inc(BufferPointer,8);
end;

procedure TStreamWriteFile.AddBuffer(var Value;NbByte:LongWord);
var InterI,NbByteToWrite:LongWord;
begin
 NbByteToWrite:=NbByte;
 while NbByteToWrite>0 do
 begin
  BufferPointer:=(BufferPointer+7) and $FFFFFFF8;
  if BufferPointer>=BufferSizeBit then
   flush;
  InterI:=min(NbByteToWrite,((BufferSizeBit-BufferPointer) shr 3));
  move(TTabBuffer(Value)[NbByte-NbByteToWrite],Buffer[BufferPointer shr 3],InterI);
  NbByteToWrite:=NbByteToWrite-InterI;
  BufferPointer:=BufferPointer+InterI shl 3;
 end; 
end;

destructor TStreamWriteFile.Destroy;
begin
 if Buffer<>Nil then
  freemem(Buffer,BufferSize);
 inherited Destroy;
end;







constructor TStreamReadFile.Create(FileName:String;BufferSize:LongWord);
begin
 inherited Create;
 Self.Filename:=Filename;
 Self.BufferSize:=BufferSize;
 Buffer:=Nil;
 FileOpened:=False;
end;
procedure TStreamReadFile.Init;
begin
 if FileOpened then
  Close;
 assignFile(F,FileName);
 try
 reset(F,1);
 FileOpened:=True;

 if Buffer=Nil then
  getmem(Buffer,BufferSize);
 BufferPointer:=0;
 BufferFilled:=0;

 except
  on E:EOutOfMemory do
   raise EFileStreamError.Create('Out of memory !');
  on E:EInOutError do
   raise EFileStreamError.Create(E.Message);
 end;
end;

procedure TStreamReadFile.Close;
begin
 if FileOpened then
 begin
  closeFile(F);
  FileOpened:=False;
 end;
end;

procedure TStreamReadFile.read;
var FileS,FileP:LongWord;
begin
 try
 BufferFilled:=0;
 if EOF(F) then
  raise EFileStreamError.Create('End of File !');
 FileP:=FilePos(F);
 FileS:=FileSize(F);
 BufferFilled:=Min(BufferSize,FileS-FileP);
 BlockRead(F,Buffer^,BufferFilled);
 BufferFilled:=BufferFilled shl 3;
 BufferPointer:=0;

 except
  on E:EInOutError do
   raise EFileStreamError.Create(E.Message);
 end;
end;
       
function TStreamReadFile.getPosition:LongWord;
var FileP:LongWord;                                                            
begin
 try
 FileP:=FilePos(F); 
 FileP:=(FileP shl 3)-BufferFilled;
 Result:=BufferPointer+FileP;
 except
  on E:EInOutError do
   raise EFileStreamError.Create(E.Message);
 end;
end;

procedure TStreamReadFile.setPosition(Position:LongWord);
begin
 try
 Seek(F,Position shr 3);
 Read;
 BufferPointer:=Position and $00000007;
 except
  on E:EInOutError do
   raise EFileStreamError.Create(E.Message);
 end;
end;

function TStreamReadFile.ReadBit:Integer;
begin
 if BufferPointer>=BufferFilled then
  read;
 result:=(Buffer[BufferPointer shr 3] shr (BufferPointer and $00000007)) and 1;
 inc(BufferPointer);
end;

function TStreamReadFile.ReadByte:Byte;
begin
 BufferPointer:=(BufferPointer+7) and $FFFFFFF8;
 if BufferPointer>=BufferFilled then
  Read;
 Result:=Buffer[BufferPointer shr 3];
 inc(BufferPointer,8);
end;

procedure TStreamReadFile.ReadBuffer(var Value;NbByte:LongWord); 
var InterI,NbByteToRead:LongWord;
begin
 NbByteToRead:=NbByte;
 while NbByteToRead>0 do
 begin
  BufferPointer:=(BufferPointer+7) and $FFFFFFF8;
  if BufferPointer>=BufferFilled then
   Read;
  InterI:=min(NbByteToRead,((BufferFilled-BufferPointer) shr 3));
  move(Buffer[BufferPointer shr 3],TTabBuffer(Value)[NbByte-NbByteToRead],InterI);
  NbByteToRead:=NbByteToRead-InterI;
  BufferPointer:=BufferPointer+InterI shl 3;
 end; 
end;


destructor TStreamReadFile.Destroy;
begin
 if Buffer<>Nil then
  freemem(Buffer,BufferSize);
 inherited Destroy;
end;





const CompressionMethod_SimpleHuffman=1;
      CompressionMethod_HuffmanDiff=2;
procedure HuffmanDiffCompression(LargX,LargY:Integer;TypeData:Integer;Image:Pointer;OutPutStream:TStreamWrite;ProcessProgress:TProcessProgress);
type TTabBuffer=array[0..MaxInt-1] of byte;
     PTabBuffer=^TTabBuffer;

     TRepresentation=record
      Value:array[0..31] of byte;
      NbBit:Byte;
     end;
     TTabRepresentation=array[0..255] of TRepresentation;

var I,J,K,InterI,InterI2:Integer;
    InterS:Single;

    BitPointer:LongWord;
    Histo:array[0..255] of LongWord;
    HistoIndex:array[0..255] of LongWord;

    TabRepresentation:TTabRepresentation;

    CompressionMethod:Integer;
    ImageSize,SizeCompressed:LongWord;

    Img_Byt:PImg_Byt;
    Img_Int:PImg_Int;
    Img_Single:PImg_Single;
    Img_Compl:PImg_Compl;

    LinePosition:PLig_LongWord;

procedure AddBit(Buffer:Pointer;BitPointer:LongWord;BitValue:Boolean);
begin
 if BitValue then
  PTabBuffer(Buffer)[BitPointer shr 3]:=PTabBuffer(Buffer)[BitPointer shr 3] or (1 shl (BitPointer and $00000007));
end;

procedure AddHuffmanValue(Value:Integer);
begin
 OutPutStream.AddArrayOfBit(TabRepresentation[Value].Value,TabRepresentation[Value].NbBit);
 inc(BitPointer,TabRepresentation[Value].NbBit);
end;

procedure CreateTree(K:LongWord;I1,I2:Integer);
var I,J:Integer;
    Sum:LongWord;
begin
 I:=I1;
 Sum:=Histo[I];
 while Sum<(K shr 1) do
 begin
  inc(I);
  Sum:=Sum+Histo[I];
 end;

 //"Arrondi"
 if I<>I1 then
  if abs(Sum-(K shr 1))>abs(Sum-Histo[I]-(K shr 1)) then
  begin
   Sum:=Sum-Histo[I];
   dec(I);
  end;

 for J:=I1 to I do
 begin
  addBit(@TabRepresentation[HistoIndex[J]].Value,TabRepresentation[HistoIndex[J]].NbBit,False);
  inc(TabRepresentation[HistoIndex[J]].NbBit);
 end;
 for J:=I+1 to I2 do
 begin
  addBit(@TabRepresentation[HistoIndex[J]].Value,TabRepresentation[HistoIndex[J]].NbBit,True);
  inc(TabRepresentation[HistoIndex[J]].NbBit);
 end;

 if I>I1 then
  CreateTree(Sum,I1,I);
 if I+1<I2 then
  CreateTree(K-Sum,I+1,I2);
end;

begin

 Img_Byt:=Nil;
 Img_Int:=Nil;
 Img_Single:=Nil;
 Img_Compl:=Nil;
 ImageSize:=0;

 case TypeData of
  2:begin
     Img_Byt:=Image;
     ImageSize:=LargX*LargY;
     SizeCompressed:=ImageSize;
    end;
  3:begin
     Img_Int:=Image;
     ImageSize:=LargX*LargY*2;
     SizeCompressed:=ImageSize;
    end;
  6:begin
     Img_Single:=Image;
     ImageSize:=LargX*LargY*4;
     SizeCompressed:=ImageSize;
    end;
 11:begin
     Img_Compl:=Image;
     ImageSize:=LargX*LargY*8;
     SizeCompressed:=ImageSize;
    end;
 end;

 //Line positions every each 8 lines
 getmem(LinePosition,((LargY shr 3)+1)*SizeOf(LongWord));
 try

 case TypeData of
  3:begin
     //Image analysis and Huffman init
     fillchar(Histo,SizeOf(Histo),0);
     SizeCompressed:=0;
     for I:=1 to LargY do
     begin
      InterI:=Img_Int[I][1];
      InterI2:=InterI+32768;
      inc(Histo[byte(InterI2)]);
      inc(Histo[byte(InterI2 shr 8)]);
      inc(SizeCompressed,2);
      for J:=2 to LargX do
      begin
       InterI2:=Img_Int[I][J]-InterI;
       InterI:=Img_Int[I][J];
       if abs(InterI2)<=127 then
       begin
        inc(Histo[InterI2+128]);
        inc(SizeCompressed);
       end
       else
       begin
        InterI2:=InterI+32768;
        inc(Histo[0]);
        inc(Histo[byte(InterI2)]);
        inc(Histo[byte(InterI2 shr 8)]);
        inc(SizeCompressed,3);
       end;
      end;
     end;
    end;
 end;


 if SizeCompressed>=ImageSize then
 begin
  CompressionMethod:=CompressionMethod_SimpleHuffman;
  SizeCompressed:=ImageSize;
  //Image analysis and Huffman init
  fillchar(Histo,SizeOf(Histo),0);
  case TypeData of
   2:begin
      for I:=1 to LargY do
       for J:=1 to LargX do
        inc(Histo[Img_Byt[I][J]]);
     end;
   3:begin
      for I:=1 to LargY do
       for J:=1 to LargX do
       begin
        InterI:=Img_Int[I][J];
        InterI:=InterI+32768;
        inc(Histo[byte(InterI)]);
        inc(Histo[byte(InterI shr 8)]);
       end;
     end;
   6:begin
      for I:=1 to LargY do
       for J:=1 to LargX do
       begin
        InterS:=Img_Single[I][J];
        asm
         mov eax,InterS
         mov InterI,eax
        end;
        inc(Histo[byte(InterI)]);
        inc(Histo[byte(InterI shr 8)]);
        inc(Histo[byte(InterI shr 16)]);
        inc(Histo[byte(InterI shr 24)]);
       end;
     end;
  11:begin
      for I:=1 to LargY do
       for J:=1 to LargX do
       begin
        InterS:=Img_Compl[I][J].Re;
        asm
         mov eax,InterS
         mov InterI,eax
        end;
        inc(Histo[byte(InterI)]);
        inc(Histo[byte(InterI shr 8)]);
        inc(Histo[byte(InterI shr 16)]);
        inc(Histo[byte(InterI shr 24)]);
        InterS:=Img_Compl[I][J].Im;
        asm
         mov eax,InterS
         mov InterI,eax
        end;
        inc(Histo[byte(InterI)]);
        inc(Histo[byte(InterI shr 8)]);
        inc(Histo[byte(InterI shr 16)]);
        inc(Histo[byte(InterI shr 24)]);
       end;
     end;
  end;
 end
 else
  CompressionMethod:=CompressionMethod_HuffmanDiff;

 //Sort
 for I:=0 to 255 do
  HistoIndex[I]:=I;
 for I:=0 to 254 do
  for J:=255 downto I+1 do
  begin
   if Histo[I]<Histo[J] then
   begin
    InterI:=Histo[I];
    Histo[I]:=Histo[J];
    Histo[J]:=InterI;
    InterI:=HistoIndex[I];
    HistoIndex[I]:=HistoIndex[J];
    HistoIndex[J]:=InterI;
   end;
  end;


 //Tree
 fillchar(TabRepresentation,SizeOf(TabRepresentation),0);
 CreateTree(SizeCompressed,0,255);


 //Compressed size calculation
 BitPointer:=256*8;
 for I:=0 to 255 do
  inc(BitPointer,TabRepresentation[HistoIndex[I]].NbBit*Histo[I]);
 for I:=0 to 255 do
  inc(BitPointer,TabRepresentation[I].NbBit);
 SizeCompressed:=((BitPointer+7) shr 3);

 OutputStream.AddByte(CompressionMethod);
 OutputStream.AddBuffer(SizeCompressed,4);


 //Copy Tree
 //Copy nbBit tab
 for I:=0 to 255 do
  OutputStream.AddByte(TabRepresentation[I].NbBit);

 //Copy bit tab
 for I:=0 to 255 do
  AddHuffmanValue(I);

 BitPointer:=0; 
 //Copy compressed data 
 if CompressionMethod=CompressionMethod_SimpleHuffman then
 begin
  case TypeData of
   2:begin
      for I:=1 to LargY do
      begin
       if assigned(ProcessProgress) then
        ProcessProgress.progress(I-1,LargY-1);

       if (((I-1) and $00000007)=0) then
        LinePosition[((I-1) shr 3)+1]:=BitPointer;

       for J:=1 to LargX do
        AddHuffmanValue(Img_Byt[I][J]);
      end;  
     end;
   3:begin
      for I:=1 to LargY do
      begin
       if assigned(ProcessProgress) then
        ProcessProgress.progress(I-1,LargY-1);
        
       if (((I-1) and $00000007)=0) then
        LinePosition[((I-1) shr 3)+1]:=BitPointer;

       for J:=1 to LargX do
       begin
        InterI:=Img_Int[I][J];
        InterI:=InterI+32768;
        AddHuffmanValue(byte(InterI));
        AddHuffmanValue(byte(InterI shr 8));
       end;
      end; 
     end;
   6:begin
      for I:=1 to LargY do
      begin
       if assigned(ProcessProgress) then
        ProcessProgress.progress(I-1,LargY-1);

       if (((I-1) and $00000007)=0) then
        LinePosition[((I-1) shr 3)+1]:=BitPointer;
        
       for J:=1 to LargX do
       begin
        InterS:=Img_Single[I][J];
        asm
         mov eax,InterS
         mov InterI,eax
        end;
        AddHuffmanValue(byte(InterI));
        AddHuffmanValue(byte(InterI shr 8));
        AddHuffmanValue(byte(InterI shr 16));
        AddHuffmanValue(byte(InterI shr 24));
       end;
      end; 
     end;
  11:begin
      for I:=1 to LargY do
      begin
       if assigned(ProcessProgress) then
        ProcessProgress.progress(I-1,LargY-1);

       if (((I-1) and $00000007)=0) then
        LinePosition[((I-1) shr 3)+1]:=BitPointer;

       for J:=1 to LargX do
       begin
        InterS:=Img_Compl[I][J].Re;
        asm
         mov eax,InterS
         mov InterI,eax
        end;
        AddHuffmanValue(byte(InterI));
        AddHuffmanValue(byte(InterI shr 8));
        AddHuffmanValue(byte(InterI shr 16));
        AddHuffmanValue(byte(InterI shr 24));
        InterS:=Img_Compl[I][J].Im;
        asm
         mov eax,InterS
         mov InterI,eax
        end;
        AddHuffmanValue(byte(InterI));
        AddHuffmanValue(byte(InterI shr 8));
        AddHuffmanValue(byte(InterI shr 16));
        AddHuffmanValue(byte(InterI shr 24));
       end;
      end; 
     end;
  end;   
 end
 else
 begin
  for I:=1 to LargY do
  begin
   if assigned(ProcessProgress) then
    ProcessProgress.progress(I-1,LargY-1);

   if (((I-1) and $00000007)=0) then
    LinePosition[((I-1) shr 3)+1]:=BitPointer;


   InterI:=Img_Int[I][1];
   InterI2:=InterI+32768;
   AddHuffmanValue(byte(InterI2));
   AddHuffmanValue(byte(InterI2 shr 8));
   for J:=2 to LargX do
   begin
    InterI2:=Img_Int[I][J]-InterI;
    InterI:=Img_Int[I][J];
    if abs(InterI2)<=127 then
     AddHuffmanValue(byte(InterI2+128))
    else
    begin
     InterI2:=InterI+32768;
     AddHuffmanValue(byte(0));
     AddHuffmanValue(byte(InterI2));
     AddHuffmanValue(byte(InterI2 shr 8));
    end;
   end;
  end;
 end;

 //Line positions
 OutputStream.AddBuffer(LinePosition^,((LargY shr 3)+1)*SizeOf(LongWord));

 finally
 freemem(LinePosition,((LargY shr 3)+1)*SizeOf(LongWord));
 end;
end;


procedure HuffmanDiffUnCompression(InputStream:TStreamRead;LargX,LargY:Integer;TypeData:Integer;Image:Pointer;X1,Y1,X2,Y2:Integer;ProcessProgress:TProcessProgress);
type PNode=^TNode;
     TNode=record
      NodeChild:array[0..1] of PNode;
      Value:Byte;
     end;
     TTree=array[0..511] of TNode;


var I,J,InterI,InterI2:Integer;
    InterLW:LongWord;
    InterS:Single;
    Tree:TTree;
    SizeCompressed:Longword;
    CompressionMethod:Integer;

    Img_Byt:PImg_Byt;
    Img_Int:PImg_Int;
    Img_Single:PImg_Single;
    Img_Compl:PImg_Compl;


    LinePosition:PLig_LongWord;

    Y1CurrentPosition:Integer;
    
    EndOfStreamPosition:LongWord;
    

procedure CreateTree;
var I,J:Integer;
    BitValue:Integer;
    SizeBitRepresentation:array[0..255] of byte;

    CurrentNode:PNode;
    LastFreeNode:Integer;
begin
 try
 for I:=0 to 255 do
  SizeBitRepresentation[I]:=InputStream.ReadByte;
 except
  on E:EFileStreamError do
   raise EDataExtractionError.create(6);
 end;

 fillchar(Tree,SizeOf(Tree),0);
 LastFreeNode:=1;
 try
 for I:=0 to 255 do
 begin
  CurrentNode:=@Tree[0];
  for J:=1 to SizeBitRepresentation[I] do
  begin
   BitValue:=InputStream.ReadBit;
   if CurrentNode.NodeChild[BitValue]=Nil then
   begin
    if LastFreeNode>=512 then
     raise EDataExtractionError.create(5);
    CurrentNode.NodeChild[BitValue]:=@Tree[LastFreeNode];
    inc(LastFreeNode);
   end;
   CurrentNode:=CurrentNode.NodeChild[BitValue];
  end;
  CurrentNode.Value:=I;
 end;
 except
  on E:EFileStreamError do
   raise EDataExtractionError.create(4);
 end;
end;


function ExtractValue:Integer;
var BitValue:Integer;
    CurrentNode:PNode;
begin
 CurrentNode:=@Tree[0];
 while (CurrentNode.NodeChild[0]<>Nil) or (CurrentNode.NodeChild[1]<>Nil) do
 begin
  BitValue:=InputStream.ReadBit;
  CurrentNode:=CurrentNode.NodeChild[BitValue];
  if CurrentNode=Nil then
   raise EDataExtractionError.create(3);
 end;
 Result:=CurrentNode.Value;
end;

begin
 Img_Byt:=Nil;
 Img_Int:=Nil;
 Img_Single:=Nil;
 Img_Compl:=Nil;

 case TypeData of
  2:Img_Byt:=Image;
  3:Img_Int:=Image;
  6:Img_Single:=Image;
 11:Img_Compl:=Image;
 end;
 CompressionMethod:=InputStream.ReadByte;
 InputStream.ReadBuffer(SizeCompressed,4);

 case TypeData of
  2:if not (CompressionMethod in [CompressionMethod_SimpleHuffman]) then
     raise EDataExtractionError.create(2);
  3:if not (CompressionMethod in [CompressionMethod_SimpleHuffman,CompressionMethod_HuffmanDiff]) then
     raise EDataExtractionError.create(2);
  6:if not (CompressionMethod in [CompressionMethod_SimpleHuffman]) then
     raise EDataExtractionError.create(2);
 11:if not (CompressionMethod in [CompressionMethod_SimpleHuffman]) then
     raise EDataExtractionError.create(2);
 end;

 try

 getmem(LinePosition,((LargY shr 3)+1)*SizeOf(LongWord));
 try

 if Y1<>1 then
 begin
  //Find the line number Y1 position
  InterLW:=InputStream.getPosition;
  InputStream.setPosition(InterLW+SizeCompressed shl 3);
  //Line positions every each 8 lines
  InterI:=((LargY shr 3)+1)*SizeOf(LongWord);
  InputStream.ReadBuffer(LinePosition^,InterI);
  InputStream.setPosition(InterLW);
 end;

 EndOfStreamPosition:=InputStream.getPosition+SizeCompressed shl 3+((((LongWord(LargY) shr 3)+1)*SizeOf(LongWord)) shl 3);
 

 
 CreateTree;
 Y1CurrentPosition:=1;
 if Y1<>1 then
 begin
  //seek the line number Y1
  InterLW:=InputStream.getPosition;
  //Line positions every each 8 lines
  InterI:=(Y1-1) and $FFFFFFF8;
  InputStream.setPosition(InterLW+LinePosition[(InterI shr 3)+1]);
  Y1CurrentPosition:=InterI;
 end;

 if CompressionMethod=CompressionMethod_SimpleHuffman then
 begin
  case TypeData of
   2:begin
      for I:=Y1CurrentPosition to Y1-1 do
       for J:=1 to LargX do
        ExtractValue;

      for I:=Y1 to Y2 do
      begin
       if assigned(ProcessProgress) then
        ProcessProgress.progress(I-1,LargY-Y1);

       for J:=1 to X1-1 do
        ExtractValue;
       for J:=X1 to X2 do
        Img_Byt[I-Y1+1][J-X1+1]:=ExtractValue;
       for J:=X2+1 to LargX do
        ExtractValue;
      end;
     end;
   3:begin
      for I:=Y1CurrentPosition to Y1-1 do
       for J:=1 to LargX do
       begin
        ExtractValue;
        ExtractValue;
       end;

      for I:=Y1 to Y2 do
      begin
       if assigned(ProcessProgress) then
        ProcessProgress.progress(I-1,LargY-Y1);

       for J:=1 to X1-1 do
       begin
        ExtractValue;
        ExtractValue;
       end;
       for J:=1 to LargX do
       begin
        InterI:=ExtractValue;
        InterI:=(InterI or (ExtractValue shl 8))-32768;
        Img_Int[I-Y1+1][J-X1+1]:=InterI;
       end;

       for J:=X2+1 to LargX do
       begin
        ExtractValue;
        ExtractValue;
       end;
      end;
     end;
   6:begin
      for I:=Y1CurrentPosition to Y1-1 do
       for J:=1 to LargX do
       begin
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
       end;

      for I:=Y1 to Y2 do
      begin
       if assigned(ProcessProgress) then
        ProcessProgress.progress(I-1,LargY-Y1);


       for J:=1 to X1-1 do
       begin
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
       end;

       for J:=1 to LargX do
       begin
        InterI:=ExtractValue;
        InterI:=InterI or (ExtractValue shl 8);
        InterI:=InterI or (ExtractValue shl 16);
        InterI:=InterI or (ExtractValue shl 24);
        asm
         mov eax,InterI
         mov InterS,eax
        end;
        Img_Single[I-Y1+1][J-X1+1]:=InterS;
       end;

       for J:=X2+1 to LargX do
       begin
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
       end;
     end;
     end;
  11:begin
      for I:=Y1CurrentPosition to Y1-1 do
       for J:=1 to LargX do
       begin
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
       end;

      for I:=Y1 to Y2 do
      begin
       if assigned(ProcessProgress) then
        ProcessProgress.progress(I-1,LargY-Y1);

       for J:=1 to X1-1 do
       begin
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
       end;

       for J:=1 to LargX do
       begin
        InterI:=ExtractValue;
        InterI:=InterI or (ExtractValue shl 8);
        InterI:=InterI or (ExtractValue shl 16);
        InterI:=InterI or (ExtractValue shl 24);
        asm
         mov eax,InterI
         mov InterS,eax
        end;
        Img_Compl[I-Y1+1][J-X1+1].Re:=InterS;
        InterI:=ExtractValue;
        InterI:=InterI or (ExtractValue shl 8);
        InterI:=InterI or (ExtractValue shl 16);
        InterI:=InterI or (ExtractValue shl 24);
        asm
         mov eax,InterI
         mov InterS,eax
        end;
        Img_Compl[I-Y1+1][J-X1+1].Im:=InterS;
       end;

       for J:=X2+1 to LargX do
       begin
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
        ExtractValue;
       end;

      end;
     end;
  end;
 end
 else
 begin
  for I:=Y1CurrentPosition to Y1-1 do
  begin
   InterI2:=ExtractValue;
   InterI:=(InterI2 or (ExtractValue shl 8))-32768;
   for J:=2 to LargX do
   begin
    InterI2:=ExtractValue;
    if InterI2<>0 then
     InterI:=InterI2-128+InterI
    else
    begin
     InterI2:=ExtractValue;
     InterI:=(InterI2 or (ExtractValue shl 8))-32768;
    end;
   end;
  end;

  for I:=Y1 to Y2 do
  begin
   if assigned(ProcessProgress) then
    ProcessProgress.progress(I-1,LargY-Y1);

   InterI2:=ExtractValue;
   InterI:=(InterI2 or (ExtractValue shl 8))-32768;

   if X1=1 then
    Img_Int[I-Y1+1][1]:=InterI;
   for J:=2 to LargX do
   begin
    InterI2:=ExtractValue;
    if InterI2<>0 then
     InterI:=InterI2-128+InterI
    else
    begin
     InterI2:=ExtractValue;
     InterI:=(InterI2 or (ExtractValue shl 8))-32768;
    end;
    if (J>=X1) and (J<=X2) then
     Img_Int[I-Y1+1][J-X1+1]:=InterI;
   end;
  end;
 end;

 //Go to the end of the file if possible  (Stream principle)
 try
 InputStream.setPosition(EndOfStreamPosition);
 except
  on E:EFileStreamError do;
 end;

 finally
 freemem(LinePosition,((LargY shr 3)+1)*SizeOf(LongWord));
 end;
 except
  on E:EFileStreamError do
   raise EDataExtractionError.create(1);
 end;


end;



const Sign_V4dCpa =  $07C1E1FD;


      TailleBufferv4c=512-sizeof(Boolean)-sizeof(double)-SizeOf(Double)-SizeOf(Boolean)-SizeOf(Double);
      NbMaxPlateCoef=20;

type
   TPlateVecteur=array[1..NbMaxPlateCoef] of double;


   TAstrometriePar=record
     XC,YC,LargX,LargY  : Integer;
     Alpha_Ref,Delta_Ref:Double;
     RayonX,RayonY:Double;
    end;

   TEnteteCPA_Ver4c = packed Record  // lZW supporte Magouille qui permet d'ajouter des variables...
      Signature      : Longint     ;
      Largeur        : Word        ;
      Longueur       : Word        ;
      BinningX       : Word        ;
      BinningY       : Word        ;
      TabSeuilHaut   : Array[1..3] of Double;
      TabSeuilBas    : Array[1..3] of Double;
      TypeData       : Byte        ;
      NbrePlan       : Byte        ;
      TimeDate       : TdateTime   ;
      TempsDePose    : Double      ;
      MiroirX        : Boolean     ;
      MiroirY        : Boolean     ;
      Telescope      : string[80]  ;
      Observateur    : string[80]  ;
      Camera         : string[80]  ;
      Filtre         : string[80]  ;
      Observatoire   : string[80]  ;
      Focale         : Double      ;
      Alpha          : Double      ;
      Delta          : Double      ;
      PixX           : Double      ;
      PixY           : Double      ;
      DebX,DebY,
      FinX,FinY      : Integer     ;
      TypeCompression: Byte        ;
      NombreBitsComp : Byte        ;

      Wavelength     : Double      ; // Longueur centrale du filtre
      Bandwidth      : Double      ; // Bande passante
      Flux           : Double      ; // Peut etre utilise pour decrire quel serait le flux une etoile de mg 0
      Diametre       : Double      ; // mm
      Offset         : Double      ; // Offest de l'inage en ADU
      FacteurCVF     : Double      ; // Facteur de convertion en e-/ADU
      Seeing         : Double      ; // en arcsec

      TemperatureCCD: Double ;
      TemperatureExt: Double ;
      Latitude,
      Longitude     : Double ;


     AstrometrieEtalon:Boolean;
     UseAstrometriePar:Boolean;
     AstrometriePar   :TAstrometriePar;
     DimPoly          :Integer ;
     SolAlpha,SolDelta:TPlateVecteur;
     SolX,SolY        :TPlateVecteur;

     TempsLectureCCD: Double ;
     Magnitude_Ref  : Double ;
     ErrFlux_Ref    : Double ;
     TDImode        : Boolean;
     TempLectureLigne: Double ;
     Commentaires   : Array [1..4] of String[ 80] ;

     // Ajouter ici et diminuer la talle de TailleBufferv4c en fonction
     // du nombre de nouvelle variables !
     // Une fois ajoutee ici
     // 1. repercuter la variable dans : TInfos_Image=record
     // 2. qu'elle est mise a jour dans la fiche de DISPLAY des infos entetes  DLGENTET.PAS
     // 3. qu'il existe une initialization dans procedure InitEntete(var Infos_Image:TInfos_Image);

     /////// NEW

     IsPhotometricCal: Boolean;
     AngleCCD        : Double ;
     Flux_Ref        : Double ;
     PhotometrieEtalonnage: Boolean;
     TransformRescaleDynamic:Double;

     Buffer:Array[1..TailleBufferv4c] of Byte; // Diminuer ici

   end;
   PEnteteCPA_ver4c=^TEnteteCPA_ver4c;



procedure FillUpInfos_Image(Entete:pointer;
                            var Infos_Image:TInfos_Image;
                            var LargX,LargY,NbrePlan,TypeData:Integer;
                            var Tab_SeuilHaut,Tab_SeuilBas:array of Extended);
var I:Integer;
    EnteteVer4c:PEnteteCPA_ver4c;
begin
 EnteteVer4c:=Entete;

 with Infos_Image do
 begin
  LargX:= EnteteVer4c.Largeur;
  LargY:= EnteteVer4c.Longueur;
  NbrePlan:=EnteteVer4c.NbrePlan;
  for I:=0 to NbrePlan-1 do
  begin
   Tab_SeuilHaut[I]:=EnteteVer4c.TabSeuilHaut[I+1];
   Tab_SeuilBas[I]:=EnteteVer4c.TabSeuilBas[I+1];
  end;
  TypeData:=EnteteVer4c.TypeData;
  BinX           :=EnteteVer4c.BinningX  ;
  BinY           :=EnteteVer4c.BinningY  ;
  Date           := Trunc(EnteteVer4c.TimeDate);
  Heure          := Frac (EnteteVer4c.TimeDate);
  Pose           := Round(EnteteVer4c.TempsDePose);
  Miroir_X       := EnteteVer4c.MiroirX;
  Miroir_Y       := EnteteVer4c.MiroirY;
  Telescope      := EnteteVer4c.Telescope;
  Observateur    := EnteteVer4c.Observateur;
  Camera         := EnteteVer4c.Camera;
  Filtre         := EnteteVer4c.Filtre;
  Observatory    := EnteteVer4c.Observatoire;
  Focale         := EnteteVer4c.Focale;
  Alpha          := EnteteVer4c.Alpha;
  Delta          := EnteteVer4c.Delta;
  PixX           := EnteteVer4c.PixX;
  PixY           := EnteteVer4c.PixY;
  DebX           := EnteteVer4c.DebX;
  DebY           := EnteteVer4c.DebY;
  FinX           := EnteteVer4c.FinX;
  FinY           := EnteteVer4c.FinY;
  Comment[1]:= EnteteVer4c.Commentaires[1];
  Comment[2]:= EnteteVer4c.Commentaires[2];
  Comment[3]:= EnteteVer4c.Commentaires[3];
  Comment[4]:= EnteteVer4c.Commentaires[4];
  Wavelength    :=EnteteVer4c.Wavelength;
  Bandwidth     :=EnteteVer4c.Bandwidth;
  Flux          :=EnteteVer4c.Flux;
  Diametre      :=EnteteVer4c.Diametre;
  Offset        :=EnteteVer4c.Offset;
  FacteurCVF    :=EnteteVer4c.FacteurCVF;
  Seeing        :=EnteteVer4c.Seeing;

  TemperatureCCD:=EnteteVer4c.TemperatureCCD;
  TemperatureExt:=EnteteVer4c.TemperatureExt;
  Latitude      :=EnteteVer4c.Latitude;
  Longitude     :=EnteteVer4c.Longitude;

  TempsLectureCCD :=EnteteVer4c.TempsLectureCCD;
  PhotometrieEtalonnage:=EnteteVer4c.PhotometrieEtalonnage;
  Magnitude_Ref   :=EnteteVer4c.Magnitude_Ref;
  Flux_Ref        :=EnteteVer4c.Flux_Ref;
  ErrFlux_Ref     :=EnteteVer4c.ErrFlux_Ref;
  TDImode         :=EnteteVer4c.TDImode                  ;
  TempLectureLigne:=EnteteVer4c.TempLectureLigne;
  PhotometricCal  :=EnteteVer4c.IsPhotometricCal;
  AngleCCD        :=EnteteVer4c.AngleCCD;
 end;
end;


procedure LoadCPA(const Name:PChar;
                  var HighValue,LowValue:TThreshold;var Infos_Image:TInfos_Image;
                  var SizeX,SizeY:Integer;var TypeData:Integer;var Image:Pointer;var NbrePlan:Integer;
                  ProcessProgress:TProcessProgress);
var StreamReadFile:TStreamReadFile;
    Signature:Integer;
    EnteteVer4c:PEnteteCPA_ver4c;

    I:Integer;
    Tab_P:array[0..2] of pointer;
begin
 StreamReadFile:=TStreamReadFile.Create(name,32*1024);
 try
 try
 StreamReadFile.Init;
 try


 StreamReadFile.ReadBuffer(Signature,4);
 if Signature=Integer(Sign_V4dCpa) then
 begin
  new(EnteteVer4c);
  try
  StreamReadFile.Init;
  StreamReadFile.ReadBuffer(EnteteVer4c^,SizeOf(TEnteteCPA_ver4c));
  FillUpInfos_Image(EnteteVer4c,Infos_Image,SizeX,SizeY,NbrePlan,TypeData,HighValue,LowValue);
  finally
  dispose(EnteteVer4c);
  end;

  //Allocate memory for image
  GetImage(SizeX,SizeY,TypeData,Image,NbrePlan);
  try

  if NbrePlan>1 then
  begin
   move(Image^,Tab_P,NbrePlan*SizeOf(Pointer));
   for I:=0 to NbrePlan-1 do
   begin
    if assigned(ProcessProgress) then
     ProcessProgress.SetProgressPonderation(1/NbrePlan,100*I/NbrePlan);
    HuffmanDiffUnCompression(StreamReadFile,SizeX,SizeY,TypeData,Tab_P[I],1,1,SizeX,SizeY,ProcessProgress);
   end;
  end
  else
   HuffmanDiffUnCompression(StreamReadFile,SizeX,SizeY,TypeData,Image,1,1,SizeX,SizeY,ProcessProgress);
  
  except
   FreeImage(SizeX,SizeY,TypeData,Image,NbrePlan);
   raise;
  end;
 end
 else
  raise EErrorLoadFile.Create(7);

 finally
 StreamReadFile.Close;
 end;
 except
  on E:EFileStreamError do
   raise EErrorLoadFile.Create(-1);
 end;
 finally
 StreamReadFile.Free;
 end;
end;








procedure FillUpEnteteCPA(Entete:PEnteteCPA_ver4c;
                          const Infos_Image:TInfos_Image;
                          LargX,LargY,NbrePlan,TypeData:Integer;
                          Tab_SeuilHaut,Tab_SeuilBas:array of Extended);
var I:Integer;                          
begin

 Fillchar(Entete^,SizeOf(TEnteteCPA_ver4c),0);

 Entete.Signature:= Sign_V4dCpa;
 Entete.Largeur        := LargX ;
 Entete.Longueur       := LargY ;
 Entete.NbrePlan       := NbrePlan;
 for I:=1 to NbrePlan do
 begin
  Entete.TabSeuilHaut[I]:=Tab_SeuilHaut[I-1];
  Entete.TabSeuilBas[I] :=Tab_SeuilBas[I-1];
 end;
 Entete.TypeData       := TypeData        ;
 Entete.BinningX       := Infos_Image.BinX;
 Entete.BinningY       := Infos_Image.BinY;
 Entete.TimeDate       := Infos_Image.Date+Infos_Image.Heure;
 Entete.TempsDePose    := Infos_Image.Pose;
 Entete.MiroirX        := Infos_Image.Miroir_X;
 Entete.MiroirY        := Infos_Image.Miroir_Y;
 Entete.Telescope      := Infos_Image.Telescope;
 Entete.Observateur    := Infos_Image.Observateur;
 Entete.Camera         := Infos_Image.Camera;
 Entete.Filtre         := Infos_Image.Filtre;
 Entete.Observatoire   := Infos_Image.Observatory;
 Entete.Focale         := Infos_Image.Focale;
 Entete.Alpha          := Infos_Image.Alpha;
 Entete.Delta          := Infos_Image.Delta;
 Entete.PixX           := Infos_Image.PixX;
 Entete.PixY           := Infos_Image.PixY;
 Entete.DebX           := Infos_Image.DebX;
 Entete.DebY           := Infos_Image.DebY;
 Entete.FinX           := Infos_Image.FinX;
 Entete.FinY           := Infos_Image.FinY;
 Entete.Commentaires[1]:= Infos_Image.Comment[1];
 Entete.Commentaires[2]:= Infos_Image.Comment[2];
 Entete.Commentaires[3]:= Infos_Image.Comment[3];
 Entete.Commentaires[4]:= Infos_Image.Comment[4];
 Entete.TempsLectureCCD:= Infos_Image.TempsLectureCCD;
 Entete.PhotometrieEtalonnage:=Infos_Image.PhotometrieEtalonnage;
 Entete.Magnitude_Ref  :=Infos_Image.Magnitude_Ref;
 Entete.Flux_Ref       :=Infos_Image.Flux_Ref;
 Entete.ErrFlux_Ref    :=Infos_Image.ErrFlux_Ref;
 Entete.Wavelength    :=Infos_Image.Wavelength;
 Entete.Bandwidth     :=Infos_Image.Bandwidth;
 Entete.Flux          :=Infos_Image.Flux;
 Entete.Diametre      :=Infos_Image.Diametre;
 Entete.Offset        :=Infos_Image.Offset;
 Entete.FacteurCVF    :=Infos_Image.FacteurCVF;
 Entete.Seeing        :=Infos_Image.Seeing;
 Entete.TemperatureCCD:=Infos_Image.TemperatureCCD;
 Entete.TemperatureExt:=Infos_Image.TemperatureExt;
 Entete.Latitude      :=Infos_Image.Latitude;
 Entete.Longitude     :=Infos_Image.Longitude;

 Entete.TDImode               :=Infos_Image.TDImode         ;
 Entete.TempLectureLigne      :=Infos_Image.TempLectureLigne;

 Entete.IsPhotometricCal  :=Infos_Image.PhotometricCal;
 Entete.AngleCCD          :=Infos_Image.AngleCCD;

end;





procedure WriteCPA(const Name:string;var SeuilHaut,SeuilBas;var Infos_Image:TInfos_Image;
                   LargX,LargY:Integer;TypeData:Integer;Image:Pointer;NbrePlan:Byte;
                   ProcessProgress:TProcessProgress);

type TLig_extended=array[0..(MaxInt div 12)-1] of extended;
     PLig_extended=^TLig_Double;

var StreamWriteFile:TStreamWriteFile;

    Entete:PEnteteCPA_ver4c;

    Tab_SeuilHaut,Tab_SeuilBas:array[1..3] of Extended;

    I:Integer;
    Tab_P:array[0..2] of pointer;
    
begin
 StreamWriteFile:=TStreamWriteFile.Create(Name,32*1024);
 try
 try
 StreamWriteFile.Init;
 try

 
 new(Entete);
 try
 move(SeuilHaut,Tab_SeuilHaut,NbrePlan*SizeOf(Extended));
 move(SeuilBas,Tab_SeuilBas,NbrePlan*SizeOf(Extended));
 FillUpEnteteCPA(Entete,
                 Infos_Image,LargX,LargY,NbrePlan,TypeData,
                 Tab_SeuilHaut,Tab_SeuilBas);


 StreamWriteFile.AddBuffer(Entete^,SizeOf(TEnteteCPA_ver4c));
 finally
 dispose(Entete);
 end;

 if NbrePlan>1 then
 begin
  move(Image^,Tab_P,NbrePlan*SizeOf(Pointer));
  for I:=0 to NbrePlan-1 do
  begin
   if assigned(ProcessProgress) then
    ProcessProgress.SetProgressPonderation(1/NbrePlan,100*I/NbrePlan);
   HuffmanDiffCompression(LargX,LargY,TypeData,Tab_P[I],StreamWriteFile,ProcessProgress);
  end;
 end
 else
  HuffmanDiffCompression(LargX,LargY,TypeData,Image,StreamWriteFile,ProcessProgress);
 finally
 StreamWriteFile.Close;
 end;
 except
  on E:EFileStreamError do
   raise EErrorSaveFile.Create(-1);
 end;
 finally
 StreamWriteFile.Free;
 end;
end;


end.
